@@ -17,8 +17,8 @@ FILE_LICENCE ( GPL2_OR_LATER );
 #include <stddef.h>
 #include <stdio.h>
 #include <ipxe/init.h>
+#include <ipxe/version.h>
 #include <usr/autoboot.h>
-#include <config/general.h>
 
 /**
  * Main entry point
@@ -31,7 +31,7 @@ __asmcall int main ( void ) {
 	initialise();
 
 	/* Some devices take an unreasonably long time to initialise */
-	printf ( PRODUCT_SHORT_NAME " initialising devices..." );
+	printf ( "%s initialising devices...", product_short_name );
 	startup();
 	printf ( "ok\n" );
 
